@@ -1,17 +1,22 @@
 /**
  * @copyright
  * ====================================================================
- * Copyright (c) 2000-2008 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  * @endcopyright
  *
@@ -24,8 +29,10 @@
 
 /* ### this should go away, but it causes too much breakage right now */
 #include <stdlib.h>
+#include <limits.h> /* for ULONG_MAX */
 
 #include <apr.h>         /* for apr_size_t, apr_int64_t, ... */
+#include <apr_version.h>
 #include <apr_errno.h>   /* for apr_status_t */
 #include <apr_pools.h>   /* for apr_pool_t */
 #include <apr_hash.h>    /* for apr_hash_t */
@@ -44,19 +51,123 @@ extern "C" {
  * @since New in 1.6.
  */
 #ifndef SVN_DEPRECATED
-#if !defined(SWIGPERL) && !defined(SWIGPYTHON) && !defined(SWIGRUBY)
-#if defined(__GNUC__) && (__GNUC__ >= 4 || (__GNUC__==3 && __GNUC_MINOR__>=1))
-#define SVN_DEPRECATED __attribute__((deprecated))
-#elif defined(_MSC_VER) && _MSC_VER >= 1300
-#define SVN_DEPRECATED __declspec(deprecated)
-#else
-#define SVN_DEPRECATED
+# if !defined(SWIGPERL) && !defined(SWIGPYTHON) && !defined(SWIGRUBY)
+#  if defined(__GNUC__) && (__GNUC__ >= 4 || (__GNUC__==3 && __GNUC_MINOR__>=1))
+#   define SVN_DEPRECATED __attribute__((deprecated))
+#  elif defined(_MSC_VER) && _MSC_VER >= 1300
+#   define SVN_DEPRECATED __declspec(deprecated)
+#  else
+#   define SVN_DEPRECATED
+#  endif
+# else
+#  define SVN_DEPRECATED
+# endif
 #endif
-#else
-#define SVN_DEPRECATED
+
+
+/** Macro used to mark experimental functions.
+ *
+ * @since New in 1.9.
+ */
+#ifndef SVN_EXPERIMENTAL
+# if !defined(SWIGPERL) && !defined(SWIGPYTHON) && !defined(SWIGRUBY)
+#  if defined(__has_attribute)
+#    if __has_attribute(__warning__)
+#      define SVN_EXPERIMENTAL __attribute__((warning("experimental function used")))
+#    else
+#      define SVN_EXPERIMENTAL
+#    endif
+#  elif !defined(__llvm__) && defined(__GNUC__) \
+      && (__GNUC__ >= 4 || (__GNUC__==3 && __GNUC_MINOR__>=1))
+#   define SVN_EXPERIMENTAL __attribute__((warning("experimental function used")))
+#  elif defined(_MSC_VER) && _MSC_VER >= 1300
+#   define SVN_EXPERIMENTAL __declspec(deprecated("experimental function used"))
+#  else
+#   define SVN_EXPERIMENTAL
+#  endif
+# else
+#  define SVN_EXPERIMENTAL
+# endif
 #endif
+
+/** Macro used to mark functions that require a final null sentinel argument.
+ *
+ * @since New in 1.9.
+ */
+#ifndef SVN_NEEDS_SENTINEL_NULL
+#  if defined(__has_attribute)
+#    if __has_attribute(__sentinel__)
+#      define SVN_NEEDS_SENTINEL_NULL __attribute__((sentinel))
+#    else
+#      define SVN_NEEDS_SENTINEL_NULL
+#    endif
+#  elif defined(__GNUC__) && (__GNUC__ >= 4)
+#    define SVN_NEEDS_SENTINEL_NULL __attribute__((sentinel))
+#  else
+#    define SVN_NEEDS_SENTINEL_NULL
+#  endif
 #endif
 
+
+/** Indicate whether the current platform supports unaligned data access.
+ *
+ * On the majority of machines running SVN (x86 / x64), unaligned access
+ * is much cheaper than repeated aligned access. Define this macro to 1
+ * on those machines.
+ * Unaligned access on other machines (e.g. IA64) will trigger memory
+ * access faults or simply misbehave.
+ *
+ * Note: Some platforms may only support unaligned access for integers
+ * (PowerPC).  As a result this macro should only be used to determine
+ * if unaligned access is supported for integers.
+ *
+ * @since New in 1.7.
+ */
+#ifndef SVN_UNALIGNED_ACCESS_IS_OK
+# if defined(_M_IX86) || defined(i386) \
+     || defined(_M_X64) || defined(__x86_64) \
+     || defined(__powerpc__) || defined(__ppc__)
+#  define SVN_UNALIGNED_ACCESS_IS_OK 1
+# else
+#  define SVN_UNALIGNED_ACCESS_IS_OK 0
+# endif
+#endif
+
+
+
+/** YABT:  Yet Another Boolean Type */
+typedef int svn_boolean_t;
+
+#ifndef TRUE
+/** uhh... true */
+#define TRUE 1
+#endif /* TRUE */
+
+#ifndef FALSE
+/** uhh... false */
+#define FALSE 0
+#endif /* FALSE */
+
+
+
+/* Declaration of a unique type, never defined, for the SVN_VA_NULL macro.
+ *
+ * NOTE: Private. Not for direct use by third-party code.
+ */
+struct svn__null_pointer_constant_stdarg_sentinel_t;
+
+/** Null pointer constant used as a sentinel in variable argument lists.
+ *
+ * Use of this macro ensures that the argument is of the correct size when a
+ * pointer is expected. (The macro @c NULL is not defined as a pointer on
+ * all systems, and the arguments to variadic functions are not converted
+ * automatically to the expected type.)
+ *
+ * @since New in 1.9.
+ */
+#define SVN_VA_NULL ((struct svn__null_pointer_constant_stdarg_sentinel_t*)0)
+/* See? (char*)NULL -- They have the same length, but the cast looks ugly. */
+
 
 
 /** Subversion error object.
@@ -66,27 +177,55 @@ extern "C" {
  */
 typedef struct svn_error_t
 {
-  /** APR error value, possibly SVN_ custom err */
+  /** APR error value; possibly an SVN_ custom error code (see
+   * `svn_error_codes.h' for a full listing).
+   */
   apr_status_t apr_err;
 
-  /** details from producer of error */
+  /** Details from the producer of error.
+   *
+   * Note that if this error was generated by Subversion's API, you'll
+   * probably want to use svn_err_best_message() to get a single
+   * descriptive string for this error chain (see the @a child member)
+   * or svn_handle_error2() to print the error chain in full.  This is
+   * because Subversion's API functions sometimes add many links to
+   * the error chain that lack details (used only to produce virtual
+   * stack traces).  (Use svn_error_purge_tracing() to remove those
+   * trace-only links from the error chain.)
+   */
   const char *message;
 
-  /** ptr to the error we "wrap" */
+  /** Pointer to the error we "wrap" (may be @c NULL).  Via this
+   * member, individual error object can be strung together into an
+   * "error chain".
+   */
   struct svn_error_t *child;
 
-  /** The pool holding this error and any child errors it wraps */
+  /** The pool in which this error object is allocated.  (If
+   * Subversion's APIs are used to manage error chains, then this pool
+   * will contain the whole error chain of which this object is a
+   * member.) */
   apr_pool_t *pool;
 
-  /** Source file where the error originated.  Only used iff @c SVN_DEBUG. */
+  /** Source file where the error originated (iff @c SVN_DEBUG;
+   * undefined otherwise).
+   */
   const char *file;
 
-  /** Source line where the error originated.  Only used iff @c SVN_DEBUG. */
+  /** Source line where the error originated (iff @c SVN_DEBUG;
+   * undefined otherwise).
+   */
   long line;
 
 } svn_error_t;
 
 
+
+/* See svn_version.h.
+   Defined here to avoid including svn_version.h from all public headers. */
+typedef struct svn_version_t svn_version_t;
+
+
 
 /** @defgroup APR_ARRAY_compat_macros APR Array Compatibility Helper Macros
  * These macros are provided by APR itself from version 1.3.
@@ -105,9 +244,67 @@ typedef struct svn_error_t
 #endif
 
 /** @} */
+
+
+
+/** @defgroup apr_hash_utilities APR Hash Table Helpers
+ * These functions enable the caller to dereference an APR hash table index
+ * without type casts or temporary variables.
+ *
+ * These functions are provided by APR itself from version 1.5.
+ * Definitions are provided here for when using older versions of APR.
+ * @{
+ */
+
+#if !APR_VERSION_AT_LEAST(1, 5, 0)
+
+/** Return the key of the hash table entry indexed by @a hi. */
+const void *
+apr_hash_this_key(apr_hash_index_t *hi);
+
+/** Return the key length of the hash table entry indexed by @a hi. */
+apr_ssize_t
+apr_hash_this_key_len(apr_hash_index_t *hi);
+
+/** Return the value of the hash table entry indexed by @a hi. */
+void *
+apr_hash_this_val(apr_hash_index_t *hi);
+
+#endif
+
+/** @} */
+
+
+
+/** On Windows, APR_STATUS_IS_ENOTDIR includes several kinds of
+ * invalid-pathname error but not ERROR_INVALID_NAME, so we include it.
+ * We also include ERROR_DIRECTORY as that was not included in apr versions
+ * before 1.4.0 and this fix is not backported */
+/* ### These fixes should go into APR. */
+#ifndef WIN32
+#define SVN__APR_STATUS_IS_ENOTDIR(s)  APR_STATUS_IS_ENOTDIR(s)
+#else
+#define SVN__APR_STATUS_IS_ENOTDIR(s)  (APR_STATUS_IS_ENOTDIR(s) \
+                      || ((s) == APR_OS_START_SYSERR + ERROR_DIRECTORY) \
+                      || ((s) == APR_OS_START_SYSERR + ERROR_INVALID_NAME))
+#endif
+
+/** On Windows, APR_STATUS_IS_EPIPE does not include ERROR_NO_DATA error.
+ * So we include it.*/
+/* ### These fixes should go into APR. */
+#ifndef WIN32
+#define SVN__APR_STATUS_IS_EPIPE(s)  APR_STATUS_IS_EPIPE(s)
+#else
+#define SVN__APR_STATUS_IS_EPIPE(s)  (APR_STATUS_IS_EPIPE(s) \
+                      || ((s) == APR_OS_START_SYSERR + ERROR_NO_DATA))
+#endif
+
+/** @} */
+
+
 
 /** The various types of nodes in the Subversion filesystem. */
-typedef enum
+typedef enum svn_node_kind_t
 {
   /** absent */
   svn_node_none,
@@ -119,7 +316,14 @@ typedef enum
   svn_node_dir,
 
   /** something's here, but we don't know what */
-  svn_node_unknown
+  svn_node_unknown,
+
+  /**
+   * symbolic link
+   * @note This value is not currently used by the public API.
+   * @since New in 1.8.
+   */
+  svn_node_symlink
 } svn_node_kind_t;
 
 /** Return a constant string expressing @a kind as an English word, e.g.,
@@ -134,13 +338,51 @@ svn_node_kind_to_word(svn_node_kind_t kind);
 
 /** Return the appropriate node_kind for @a word.  @a word is as
  * returned from svn_node_kind_to_word().  If @a word does not
- * represent a recognized kind or is @c NULL, return @c svn_node_unknown.
+ * represent a recognized kind or is @c NULL, return #svn_node_unknown.
  *
  * @since New in 1.6.
  */
 svn_node_kind_t
 svn_node_kind_from_word(const char *word);
 
+
+/** Generic three-state property to represent an unknown value for values
+ * that are just like booleans.  The values have been set deliberately to
+ * make tristates disjoint from #svn_boolean_t.
+ *
+ * @note It is unsafe to use apr_pcalloc() to allocate these, since '0' is
+ * not a valid value.
+ *
+ * @since New in 1.7. */
+typedef enum svn_tristate_t
+{
+  /** state known to be false (the constant does not evaulate to false) */
+  svn_tristate_false = 2,
+  /** state known to be true */
+  svn_tristate_true,
+  /** state could be true or false */
+  svn_tristate_unknown
+} svn_tristate_t;
+
+/** Return a constant string "true", "false" or NULL representing the value of
+ * @a tristate.
+ *
+ * @since New in 1.7.
+ */
+const char *
+svn_tristate__to_word(svn_tristate_t tristate);
+
+/** Return the appropriate tristate for @a word. If @a word is "true", returns
+ * #svn_tristate_true; if @a word is "false", returns #svn_tristate_false,
+ * for all other values (including NULL) returns #svn_tristate_unknown.
+ *
+ * @since New in 1.7.
+ */
+svn_tristate_t
+svn_tristate__from_word(const char * word);
+
+
+
 /** About Special Files in Subversion
  *
  * Subversion denotes files that cannot be portably created or
@@ -178,6 +420,8 @@ svn_node_kind_from_word(const char *word);
  *     routines from 1.
  */
 
+
+
 /** A revision number. */
 typedef long int svn_revnum_t;
 
@@ -189,7 +433,7 @@ typedef long int svn_revnum_t;
 
 /** Not really invalid...just unimportant -- one day, this can be its
  * own unique value, for now, just make it the same as
- * @c SVN_INVALID_REVNUM.
+ * #SVN_INVALID_REVNUM.
  */
 #define SVN_IGNORED_REVNUM ((svn_revnum_t) -1)
 
@@ -201,7 +445,7 @@ typedef long int svn_revnum_t;
  * store its value in @a rev.  If @a endptr is non-NULL, then the
  * address of the first non-numeric character in @a str is stored in
  * it.  If there are no digits in @a str, then @a endptr is set (if
- * non-NULL), and the error @c SVN_ERR_REVNUM_PARSE_FAILURE error is
+ * non-NULL), and the error #SVN_ERR_REVNUM_PARSE_FAILURE error is
  * returned.  Negative numbers parsed from @a str are considered
  * invalid, and result in the same error.
  *
@@ -224,6 +468,7 @@ svn_revnum_parse(svn_revnum_t *rev,
 #define SVN_REVNUM_T_FMT "ld"
 
 
+
 /** The size of a file in the Subversion FS. */
 typedef apr_int64_t svn_filesize_t;
 
@@ -241,20 +486,7 @@ typedef apr_int64_t svn_filesize_t;
 #endif
 
 
-/** YABT:  Yet Another Boolean Type */
-typedef int svn_boolean_t;
-
-#ifndef TRUE
-/** uhh... true */
-#define TRUE 1
-#endif /* TRUE */
-
-#ifndef FALSE
-/** uhh... false */
-#define FALSE 0
-#endif /* FALSE */
-
-
+
 /** An enum to indicate whether recursion is needed. */
 enum svn_recurse_kind
 {
@@ -269,50 +501,49 @@ enum svn_recurse_kind
  *
  * @since New in 1.5.
  */
-typedef enum
+typedef enum svn_depth_t
 {
   /* The order of these depths is important: the higher the number,
      the deeper it descends.  This allows us to compare two depths
      numerically to decide which should govern. */
 
-  /* Depth undetermined or ignored.  In some contexts, this means the
-     client should choose an appropriate default depth.  The server
-     will generally treat it as @c svn_depth_infinity. */
+  /** Depth undetermined or ignored.  In some contexts, this means the
+      client should choose an appropriate default depth.  The server
+      will generally treat it as #svn_depth_infinity. */
   svn_depth_unknown    = -2,
 
-  /* Exclude (i.e., don't descend into) directory D. */
-  /* NOTE: In Subversion 1.5, svn_depth_exclude is *not* supported
-     anywhere in the client-side (libsvn_wc/libsvn_client/etc) code;
-     it is only supported as an argument to set_path functions in the
-     ra and repos reporters.  (This will enable future versions of
-     Subversion to run updates, etc, against 1.5 servers with proper
-     svn_depth_exclude behavior, once we get a chance to implement
-     client-side support for svn_depth_exclude.)
+  /** Exclude (i.e., don't descend into) directory D.
+      @note In Subversion 1.5, svn_depth_exclude is *not* supported
+      anywhere in the client-side (libsvn_wc/libsvn_client/etc) code;
+      it is only supported as an argument to set_path functions in the
+      ra and repos reporters.  (This will enable future versions of
+      Subversion to run updates, etc, against 1.5 servers with proper
+      svn_depth_exclude behavior, once we get a chance to implement
+      client-side support for svn_depth_exclude.)
   */
   svn_depth_exclude    = -1,
 
-  /* Just the named directory D, no entries.  Updates will not pull in
-     any files or subdirectories not already present. */
+  /** Just the named directory D, no entries.  Updates will not pull in
+      any files or subdirectories not already present. */
   svn_depth_empty      =  0,
 
-  /* D + its file children, but not subdirs.  Updates will pull in any
-     files not already present, but not subdirectories. */
+  /** D + its file children, but not subdirs.  Updates will pull in any
+      files not already present, but not subdirectories. */
   svn_depth_files      =  1,
 
-  /* D + immediate children (D and its entries).  Updates will pull in
-     any files or subdirectories not already present; those
-     subdirectories' this_dir entries will have depth-empty. */
+  /** D + immediate children (D and its entries).  Updates will pull in
+      any files or subdirectories not already present; those
+      subdirectories' this_dir entries will have depth-empty. */
   svn_depth_immediates =  2,
 
-  /* D + all descendants (full recursion from D).  Updates will pull
-     in any files or subdirectories not already present; those
-     subdirectories' this_dir entries will have depth-infinity.
-     Equivalent to the pre-1.5 default update behavior. */
+  /** D + all descendants (full recursion from D).  Updates will pull
+      in any files or subdirectories not already present; those
+      subdirectories' this_dir entries will have depth-infinity.
+      Equivalent to the pre-1.5 default update behavior. */
   svn_depth_infinity   =  3
 
 } svn_depth_t;
 
-
 /** Return a constant string expressing @a depth as an English word,
  * e.g., "infinity", "immediates", etc.  The string is not localized,
  * as it may be used for client<->server communications.
@@ -322,19 +553,17 @@ typedef enum
 const char *
 svn_depth_to_word(svn_depth_t depth);
 
-
 /** Return the appropriate depth for @a depth_str.  @a word is as
  * returned from svn_depth_to_word().  If @a depth_str does not
- * represent a recognized depth, return @c svn_depth_unknown.
+ * represent a recognized depth, return #svn_depth_unknown.
  *
  * @since New in 1.5.
  */
 svn_depth_t
 svn_depth_from_word(const char *word);
 
-
-/* Return @c svn_depth_infinity if boolean @a recurse is TRUE, else
- * return @c svn_depth_files.
+/** Return #svn_depth_infinity if boolean @a recurse is TRUE, else
+ * return #svn_depth_files.
  *
  * @note New code should never need to use this, it is called only
  * from pre-depth APIs, for compatibility.
@@ -344,9 +573,8 @@ svn_depth_from_word(const char *word);
 #define SVN_DEPTH_INFINITY_OR_FILES(recurse) \
   ((recurse) ? svn_depth_infinity : svn_depth_files)
 
-
-/* Return @c svn_depth_infinity if boolean @a recurse is TRUE, else
- * return @c svn_depth_immediates.
+/** Return #svn_depth_infinity if boolean @a recurse is TRUE, else
+ * return #svn_depth_immediates.
  *
  * @note New code should never need to use this, it is called only
  * from pre-depth APIs, for compatibility.
@@ -356,9 +584,8 @@ svn_depth_from_word(const char *word);
 #define SVN_DEPTH_INFINITY_OR_IMMEDIATES(recurse) \
   ((recurse) ? svn_depth_infinity : svn_depth_immediates)
 
-
-/* Return @c svn_depth_infinity if boolean @a recurse is TRUE, else
- * return @c svn_depth_empty.
+/** Return #svn_depth_infinity if boolean @a recurse is TRUE, else
+ * return #svn_depth_empty.
  *
  * @note New code should never need to use this, it is called only
  * from pre-depth APIs, for compatibility.
@@ -368,21 +595,20 @@ svn_depth_from_word(const char *word);
 #define SVN_DEPTH_INFINITY_OR_EMPTY(recurse) \
   ((recurse) ? svn_depth_infinity : svn_depth_empty)
 
-
-/* Return a recursion boolean based on @a depth.
+/** Return a recursion boolean based on @a depth.
  *
  * Although much code has been converted to use depth, some code still
  * takes a recurse boolean.  In most cases, it makes sense to treat
  * unknown or infinite depth as recursive, and any other depth as
- * non-recursive (which in turn usually translates to @c svn_depth_files).
+ * non-recursive (which in turn usually translates to #svn_depth_files).
  */
 #define SVN_DEPTH_IS_RECURSIVE(depth)                              \
-  (((depth) == svn_depth_infinity || (depth) == svn_depth_unknown) \
-   ? TRUE : FALSE)
+  ((depth) == svn_depth_infinity || (depth) == svn_depth_unknown)
 
 
+
 /**
- * It is sometimes convenient to indicate which parts of an @c svn_dirent_t
+ * It is sometimes convenient to indicate which parts of an #svn_dirent_t
  * object you are actually interested in, so that calculating and sending
  * the data corresponding to the other fields can be avoided.  These values
  * can be used for that purpose.
@@ -414,7 +640,13 @@ svn_depth_from_word(const char *word);
 
 /** @} */
 
-/** A general subversion directory entry. */
+/** A general subversion directory entry.
+ *
+ * @note To allow for extending the #svn_dirent_t structure in future
+ * releases, always use svn_dirent_create() to allocate the stucture.
+ *
+ * @since New in 1.6.
+ */
 typedef struct svn_dirent_t
 {
   /** node kind */
@@ -438,7 +670,6 @@ typedef struct svn_dirent_t
   /* IMPORTANT: If you extend this struct, check svn_dirent_dup(). */
 } svn_dirent_t;
 
-
 /** Return a deep copy of @a dirent, allocated in @a pool.
  *
  * @since New in 1.4.
@@ -447,8 +678,16 @@ svn_dirent_t *
 svn_dirent_dup(const svn_dirent_t *dirent,
                apr_pool_t *pool);
 
-
+/**
+ * Create a new svn_dirent_t instance with all values initialized to their
+ * not-available values.
+ *
+ * @since New in 1.8.
+ */
+svn_dirent_t *
+svn_dirent_create(apr_pool_t *result_pool);
 
+
 /** Keyword substitution.
  *
  * All the keywords Subversion recognizes.
@@ -525,6 +764,7 @@ svn_dirent_dup(const svn_dirent_t *dirent,
 
 /** @} */
 
+
 
 /** All information about a commit.
  *
@@ -547,17 +787,20 @@ typedef struct svn_commit_info_t
   /** error message from post-commit hook, or NULL. */
   const char *post_commit_err;
 
+  /** repository root, may be @c NULL if unknown.
+      @since New in 1.7. */
+  const char *repos_root;
+
 } svn_commit_info_t;
 
-
 /**
- * Allocate an object of type @c svn_commit_info_t in @a pool and
+ * Allocate an object of type #svn_commit_info_t in @a pool and
  * return it.
  *
- * The @c revision field of the new struct is set to @c
- * SVN_INVALID_REVNUM.  All other fields are initialized to @c NULL.
+ * The @c revision field of the new struct is set to #SVN_INVALID_REVNUM.
+ * All other fields are initialized to @c NULL.
  *
- * @note Any object of the type @c svn_commit_info_t should
+ * @note Any object of the type #svn_commit_info_t should
  * be created using this function.
  * This is to provide for extending the svn_commit_info_t in
  * the future.
@@ -567,7 +810,6 @@ typedef struct svn_commit_info_t
 svn_commit_info_t *
 svn_create_commit_info(apr_pool_t *pool);
 
-
 /**
  * Return a deep copy @a src_commit_info allocated in @a pool.
  *
@@ -577,11 +819,12 @@ svn_commit_info_t *
 svn_commit_info_dup(const svn_commit_info_t *src_commit_info,
                     apr_pool_t *pool);
 
+
 
 /**
  * A structure to represent a path that changed for a log entry.
  *
- * @note To allow for extending the @c svn_log_changed_path2_t structure in
+ * @note To allow for extending the #svn_log_changed_path2_t structure in
  * future releases, always use svn_log_changed_path2_create() to allocate
  * the structure.
  *
@@ -601,16 +844,24 @@ typedef struct svn_log_changed_path2_t
   /** The type of the node, may be svn_node_unknown. */
   svn_node_kind_t node_kind;
 
+  /** Is the text modified, may be svn_tristate_unknown.
+   * @since New in 1.7. */
+  svn_tristate_t text_modified;
+
+  /** Are properties modified, may be svn_tristate_unknown.
+   * @since New in 1.7. */
+  svn_tristate_t props_modified;
+
   /* NOTE: Add new fields at the end to preserve binary compatibility.
      Also, if you add fields here, you have to update
      svn_log_changed_path2_dup(). */
 } svn_log_changed_path2_t;
 
 /**
- * Returns an @c svn_log_changed_path2_t, allocated in @a pool with all fields
+ * Returns an #svn_log_changed_path2_t, allocated in @a pool with all fields
  * initialized to NULL, None or empty values.
  *
- * @note To allow for extending the @c svn_log_changed_path2_t structure in
+ * @note To allow for extending the #svn_log_changed_path2_t structure in
  * future releases, this function should always be used to allocate the
  * structure.
  *
@@ -630,7 +881,7 @@ svn_log_changed_path2_dup(const svn_log_changed_path2_t *changed_path,
 
 /**
  * A structure to represent a path that changed for a log entry.  Same as
- * @c svn_log_changed_path2_t, but without the node kind.
+ * the first three fields of #svn_log_changed_path2_t.
  *
  * @deprecated Provided for backward compatibility with the 1.5 API.
  */
@@ -647,7 +898,6 @@ typedef struct svn_log_changed_path_t
 
 } svn_log_changed_path_t;
 
-
 /**
  * Return a deep copy of @a changed_path, allocated in @a pool.
  *
@@ -662,7 +912,7 @@ svn_log_changed_path_dup(const svn_log_changed_path_t *changed_path,
 /**
  * A structure to represent all the information about a particular log entry.
  *
- * @note To allow for extending the @c svn_log_entry_t structure in future
+ * @note To allow for extending the #svn_log_entry_t structure in future
  * releases, always use svn_log_entry_create() to allocate the structure.
  *
  * @since New in 1.5.
@@ -670,10 +920,10 @@ svn_log_changed_path_dup(const svn_log_changed_path_t *changed_path,
 typedef struct svn_log_entry_t
 {
   /** A hash containing as keys every path committed in @a revision; the
-   * values are (@c svn_log_changed_path_t *) stuctures.
+   * values are (#svn_log_changed_path_t *) structures.
    *
    * The subversion core libraries will always set this field to the same
-   * value as changed_paths2 for compatibity reasons.
+   * value as changed_paths2 for compatibility reasons.
    *
    * @deprecated Provided for backward compatibility with the 1.5 API.
    */
@@ -683,7 +933,8 @@ typedef struct svn_log_entry_t
   svn_revnum_t revision;
 
   /** The hash of requested revision properties, which may be NULL if it
-   * would contain no revprops. */
+   * would contain no revprops.  Maps (const char *) property name to
+   * (svn_string_t *) property value. */
   apr_hash_t *revprops;
 
   /**
@@ -700,34 +951,56 @@ typedef struct svn_log_entry_t
    * HAS_CHILDREN flag is always FALSE.
    *
    * For more information see:
-   * http://subversion.tigris.org/merge-tracking/design.html#commutative-reporting
+   * https://svn.apache.org/repos/asf/subversion/trunk/notes/merge-tracking/design.html#commutative-reporting
    */
   svn_boolean_t has_children;
 
   /** A hash containing as keys every path committed in @a revision; the
-   * values are (@c svn_log_changed_path2_t *) stuctures.
+   * values are (#svn_log_changed_path2_t *) structures.
    *
    * If this value is not @c NULL, it MUST have the same value as
    * changed_paths or svn_log_entry_dup() will not create an identical copy.
    *
    * The subversion core libraries will always set this field to the same
-   * value as changed_paths for compatibity with users assuming an older
+   * value as changed_paths for compatibility with users assuming an older
    * version.
    *
+   * @note See http://svn.haxx.se/dev/archive-2010-08/0362.shtml for
+   * further explanation.
+   *
    * @since New in 1.6.
    */
   apr_hash_t *changed_paths2;
 
+  /**
+   * Whether @a revision should be interpreted as non-inheritable in the
+   * same sense of #svn_merge_range_t.
+   *
+   * Only set when this #svn_log_entry_t instance is returned by the
+   * libsvn_client mergeinfo apis. Currently always FALSE when the
+   * #svn_log_entry_t instance is reported by the ra layer.
+   *
+   * @since New in 1.7.
+   */
+  svn_boolean_t non_inheritable;
+
+  /**
+   * Whether @a revision is a merged revision resulting from a reverse merge.
+   *
+   * @since New in 1.7.
+   */
+  svn_boolean_t subtractive_merge;
+
   /* NOTE: Add new fields at the end to preserve binary compatibility.
      Also, if you add fields here, you have to update
      svn_log_entry_dup(). */
 } svn_log_entry_t;
 
 /**
- * Returns an @c svn_log_entry_t, allocated in @a pool with all fields
+ * Returns an #svn_log_entry_t, allocated in @a pool with all fields
  * initialized to NULL values.
  *
- * @note To allow for extending the @c svn_log_entry_t structure in future
+ * @note To allow for extending the #svn_log_entry_t structure in future
  * releases, this function should always be used to allocate the structure.
  *
  * @since New in 1.5.
@@ -744,15 +1017,15 @@ svn_log_entry_create(apr_pool_t *pool);
  * @since New in 1.6.
  */
 svn_log_entry_t *
-svn_log_entry_dup(svn_log_entry_t *log_entry, apr_pool_t *pool);
+svn_log_entry_dup(const svn_log_entry_t *log_entry, apr_pool_t *pool);
 
 /** The callback invoked by log message loopers, such as
- * @c svn_ra_plugin_t.get_log() and svn_repos_get_logs().
+ * #svn_ra_plugin_t.get_log() and svn_repos_get_logs().
  *
  * This function is invoked once on each log message, in the order
  * determined by the caller (see above-mentioned functions).
  *
- * @a baton is what you think it is, and @a log_entry contains relevent
+ * @a baton is what you think it is, and @a log_entry contains relevant
  * information for the log message.  Any of @a log_entry->author,
  * @a log_entry->date, or @a log_entry->message may be @c NULL.
  *
@@ -762,7 +1035,7 @@ svn_log_entry_dup(svn_log_entry_t *log_entry, apr_pool_t *pool);
  *
  * If @a log_entry->changed_paths is non-@c NULL, then it contains as keys
  * every path committed in @a log_entry->revision; the values are
- * (@c svn_log_changed_path_t *) structures.
+ * (#svn_log_changed_path_t *) structures.
  *
  * If @a log_entry->has_children is @c TRUE, the message will be followed
  * immediately by any number of merged revisions (child messages), which are
@@ -778,26 +1051,25 @@ svn_log_entry_dup(svn_log_entry_t *log_entry, apr_pool_t *pool);
  *
  * @since New in 1.5.
  */
-
-typedef svn_error_t *(*svn_log_entry_receiver_t)
-  (void *baton,
-   svn_log_entry_t *log_entry,
-   apr_pool_t *pool);
+typedef svn_error_t *(*svn_log_entry_receiver_t)(
+  void *baton,
+  svn_log_entry_t *log_entry,
+  apr_pool_t *pool);
 
 /**
- * Similar to @c svn_log_entry_receiver_t, except this uses separate
+ * Similar to #svn_log_entry_receiver_t, except this uses separate
  * parameters for each part of the log entry.
  *
  * @deprecated Provided for backward compatibility with the 1.4 API.
  */
-typedef svn_error_t *(*svn_log_message_receiver_t)
-  (void *baton,
-   apr_hash_t *changed_paths,
-   svn_revnum_t revision,
-   const char *author,
-   const char *date,  /* use svn_time_from_cstring() if need apr_time_t */
-   const char *message,
-   apr_pool_t *pool);
+typedef svn_error_t *(*svn_log_message_receiver_t)(
+  void *baton,
+  apr_hash_t *changed_paths,
+  svn_revnum_t revision,
+  const char *author,
+  const char *date,  /* use svn_time_from_cstring() if need apr_time_t */
+  const char *message,
+  apr_pool_t *pool);
 
 
 /** Callback function type for commits.
@@ -808,21 +1080,22 @@ typedef svn_error_t *(*svn_log_message_receiver_t)
  *
  * @since New in 1.4.
  */
-typedef svn_error_t *(*svn_commit_callback2_t)
-  (const svn_commit_info_t *commit_info,
-   void *baton,
-   apr_pool_t *pool);
+typedef svn_error_t *(*svn_commit_callback2_t)(
+  const svn_commit_info_t *commit_info,
+  void *baton,
+  apr_pool_t *pool);
 
-/** Same as @c svn_commit_callback2_t, but uses individual
- * data elements instead of the @c svn_commit_info_t structure
+/** Same as #svn_commit_callback2_t, but uses individual
+ * data elements instead of the #svn_commit_info_t structure
  *
  * @deprecated Provided for backward compatibility with the 1.3 API.
  */
-typedef svn_error_t *(*svn_commit_callback_t)
-  (svn_revnum_t new_revision,
-   const char *date,
-   const char *author,
-   void *baton);
+typedef svn_error_t *(*svn_commit_callback_t)(
+  svn_revnum_t new_revision,
+  const char *date,
+  const char *author,
+  void *baton);
+
 
 
 /** A buffer size that may be used when processing a stream of data.
@@ -863,7 +1136,7 @@ typedef svn_error_t *(*svn_commit_callback_t)
 /** Validate @a mime_type.
  *
  * If @a mime_type does not contain a "/", or ends with non-alphanumeric
- * data, return @c SVN_ERR_BAD_MIME_TYPE, else return success.
+ * data, return #SVN_ERR_BAD_MIME_TYPE, else return success.
  *
  * Use @a pool only to find error allocation.
  *
@@ -876,7 +1149,6 @@ svn_error_t *
 svn_mime_type_validate(const char *mime_type,
                        apr_pool_t *pool);
 
-
 /** Return FALSE iff @a mime_type is a textual type.
  *
  * All mime types that start with "text/" are textual, plus some special
@@ -889,8 +1161,8 @@ svn_mime_type_is_binary(const char *mime_type);
 
 /** A user defined callback that subversion will call with a user defined
  * baton to see if the current operation should be continued.  If the operation
- * should continue, the function should return @c SVN_NO_ERROR, if not, it
- * should return @c SVN_ERR_CANCELLED.
+ * should continue, the function should return #SVN_NO_ERROR, if not, it
+ * should return #SVN_ERR_CANCELLED.
  */
 typedef svn_error_t *(*svn_cancel_func_t)(void *cancel_baton);
 
@@ -928,10 +1200,10 @@ typedef struct svn_lock_t
 } svn_lock_t;
 
 /**
- * Returns an @c svn_lock_t, allocated in @a pool with all fields initialized
+ * Returns an #svn_lock_t, allocated in @a pool with all fields initialized
  * to NULL values.
  *
- * @note To allow for extending the @c svn_lock_t structure in the future
+ * @note To allow for extending the #svn_lock_t structure in the future
  * releases, this function should always be used to allocate the structure.
  *
  * @since New in 1.2.
@@ -947,6 +1219,8 @@ svn_lock_create(apr_pool_t *pool);
 svn_lock_t *
 svn_lock_dup(const svn_lock_t *lock, apr_pool_t *pool);
 
+
+
 /**
  * Return a formatted Universal Unique IDentifier (UUID) string.
  *
@@ -955,6 +1229,8 @@ svn_lock_dup(const svn_lock_t *lock, apr_pool_t *pool);
 const char *
 svn_uuid_generate(apr_pool_t *pool);
 
+
+
 /**
  * Mergeinfo representing a merge of a range of revisions.
  *
@@ -984,7 +1260,7 @@ typedef struct svn_merge_range_t
  * @since New in 1.5.
  */
 svn_merge_range_t *
-svn_merge_range_dup(svn_merge_range_t *range, apr_pool_t *pool);
+svn_merge_range_dup(const svn_merge_range_t *range, apr_pool_t *pool);
 
 /**
  * Returns true if the changeset committed in revision @a rev is one
@@ -993,7 +1269,7 @@ svn_merge_range_dup(svn_merge_range_t *range, apr_pool_t *pool);
  * @since New in 1.5.
  */
 svn_boolean_t
-svn_merge_range_contains_rev(svn_merge_range_t *range, svn_revnum_t rev);
+svn_merge_range_contains_rev(const svn_merge_range_t *range, svn_revnum_t rev);
 
 
 
@@ -1001,7 +1277,7 @@ svn_merge_range_contains_rev(svn_merge_range_t *range, svn_revnum_t rev);
  *  @{ */
 
 /**
- * A representation of a segment of a object's version history with an
+ * A representation of a segment of an object's version history with an
  * emphasis on the object's location in the repository as of various
  * revisions.
  *
@@ -1010,7 +1286,7 @@ svn_merge_range_contains_rev(svn_merge_range_t *range, svn_revnum_t rev);
 typedef struct svn_location_segment_t
 {
   /** The beginning (oldest) and ending (youngest) revisions for this
-      segment. */
+      segment, both inclusive. */
   svn_revnum_t range_start;
   svn_revnum_t range_end;
 
@@ -1020,18 +1296,16 @@ typedef struct svn_location_segment_t
 
 } svn_location_segment_t;
 
-
 /**
- * A callback invoked by generators of @c svn_location_segment_t
+ * A callback invoked by generators of #svn_location_segment_t
  * objects, used to report information about a versioned object's
  * history in terms of its location in the repository filesystem over
  * time.
  */
-typedef svn_error_t *(*svn_location_segment_receiver_t)
-  (svn_location_segment_t *segment,
-   void *baton,
-   apr_pool_t *pool);
-
+typedef svn_error_t *(*svn_location_segment_receiver_t)(
+  svn_location_segment_t *segment,
+  void *baton,
+  apr_pool_t *pool);
 
 /**
  * Return a deep copy of @a segment, allocated in @a pool.
@@ -1039,11 +1313,27 @@ typedef svn_error_t *(*svn_location_segment_receiver_t)
  * @since New in 1.5.
  */
 svn_location_segment_t *
-svn_location_segment_dup(svn_location_segment_t *segment,
+svn_location_segment_dup(const svn_location_segment_t *segment,
                          apr_pool_t *pool);
+
 /** @} */
 
 
+
+/** A line number, such as in a file or a stream.
+ *
+ * @since New in 1.7.
+ */
+typedef unsigned long svn_linenum_t;
+
+/** The maximum value of an svn_linenum_t.
+ *
+ * @since New in 1.7.
+ */
+#define SVN_LINENUM_MAX_VALUE ULONG_MAX
+
+
+
 #ifdef __cplusplus
 }
 #endif /* __cplusplus */
@@ -1063,4 +1353,14 @@ svn_location_segment_dup(svn_location_segment_t *segment,
 #include "svn_error.h"
 
 
+/*
+ * Subversion developers may want to use some additional debugging facilities
+ * while working on the code. We'll pull that in here, so individual source
+ * files don't have to include this header manually.
+ */
+#ifdef SVN_DEBUG
+#include "private/svn_debug.h"
+#endif
+
+
 #endif /* SVN_TYPES_H */
