@@ -2,33 +2,40 @@
  * compat.c:  compatibility compliance logic
  *
  * ====================================================================
- * Copyright (c) 2004-2007 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
-/* ==================================================================== */
-
-/*** Includes. ***/
 #include <apr_pools.h>
-#include <assert.h>
 
+#include "svn_hash.h"
 #include "svn_error.h"
 #include "svn_pools.h"
 #include "svn_sorts.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_ra.h"
 #include "svn_io.h"
 #include "svn_compat.h"
+#include "svn_props.h"
+
+#include "private/svn_fspath.h"
+#include "private/svn_sorts_private.h"
 #include "ra_loader.h"
 #include "svn_private_config.h"
 
@@ -74,7 +81,7 @@ prev_log_path(const char **prev_path_p,
   const char *prev_path = NULL;
 
   /* It's impossible to find the predecessor path of a NULL path. */
-  assert(path);
+  SVN_ERR_ASSERT(path);
 
   /* Initialize our return values for the action and copyfrom_rev in
      case we have an unhandled case later on. */
@@ -86,7 +93,7 @@ prev_log_path(const char **prev_path_p,
   if (changed_paths)
     {
       /* See if PATH was explicitly changed in this revision. */
-      change = apr_hash_get(changed_paths, path, APR_HASH_KEY_STRING);
+      change = svn_hash_gets(changed_paths, path);
       if (change)
         {
           /* If PATH was not newly added in this revision, then it may or may
@@ -105,7 +112,7 @@ prev_log_path(const char **prev_path_p,
                 prev_path = apr_pstrdup(pool, change->copyfrom_path);
               else
                 prev_path = NULL;
-              
+
               *prev_path_p = prev_path;
               if (action_p)
                 *action_p = change->action;
@@ -114,7 +121,7 @@ prev_log_path(const char **prev_path_p,
               return SVN_NO_ERROR;
             }
         }
-      
+
       if (apr_hash_count(changed_paths))
         {
           /* The path was not explicitly changed in this revision.  The
@@ -124,11 +131,11 @@ prev_log_path(const char **prev_path_p,
              that directory's copyfrom_path. */
           int i;
           apr_array_header_t *paths;
-          
+
           /* Build a sorted list of the changed paths. */
           paths = svn_sort__hash(changed_paths,
                                  svn_sort_compare_items_as_paths, pool);
-          
+
           /* Now, walk the list of paths backwards, looking a parent of
              our path that has copyfrom information. */
           for (i = paths->nelts; i > 0; i--)
@@ -136,28 +143,28 @@ prev_log_path(const char **prev_path_p,
               svn_sort__item_t item = APR_ARRAY_IDX(paths,
                                                     i - 1, svn_sort__item_t);
               const char *ch_path = item.key;
-              int len = strlen(ch_path);
-              
+              size_t len = strlen(ch_path);
+
               /* See if our path is the child of this change path.  If
                  not, keep looking.  */
               if (! ((strncmp(ch_path, path, len) == 0) && (path[len] == '/')))
                 continue;
-              
+
               /* Okay, our path *is* a child of this change path.  If
                  this change was copied, we just need to apply the
                  portion of our path that is relative to this change's
                  path, to the change's copyfrom path.  Otherwise, this
                  change isn't really interesting to us, and our search
                  continues. */
-              change = apr_hash_get(changed_paths, ch_path, len);
+              change = item.value;
               if (change->copyfrom_path)
                 {
                   if (action_p)
                     *action_p = change->action;
                   if (copyfrom_rev_p)
                     *copyfrom_rev_p = change->copyfrom_rev;
-                  prev_path = svn_path_join(change->copyfrom_path,
-                                            path + len + 1, pool);
+                  prev_path = svn_fspath__join(change->copyfrom_path,
+                                               path + len + 1, pool);
                   break;
                 }
             }
@@ -175,7 +182,7 @@ prev_log_path(const char **prev_path_p,
         return svn_error_createf(SVN_ERR_CLIENT_UNRELATED_RESOURCES, NULL,
                                  _("Missing changed-path information for "
                                    "'%s' in revision %ld"),
-                                 svn_path_local_style(path, pool), revision);
+                                 svn_dirent_local_style(path, pool), revision);
     }
 
   *prev_path_p = prev_path;
@@ -183,6 +190,26 @@ prev_log_path(const char **prev_path_p,
 }
 
 
+/* Set *FS_PATH_P to the absolute filesystem path associated with the
+   URL built from SESSION's URL and REL_PATH (which is relative to
+   session's URL.  Use POOL for allocations. */
+static svn_error_t *
+get_fs_path(const char **fs_path_p,
+            svn_ra_session_t *session,
+            const char *rel_path,
+            apr_pool_t *pool)
+{
+  const char *url, *fs_path;
+
+  SVN_ERR(svn_ra_get_session_url(session, &url, pool));
+  SVN_ERR(svn_ra_get_path_relative_to_root(session, &fs_path, url, pool));
+  *fs_path_p = svn_fspath__canonicalize(svn_relpath_join(fs_path,
+                                                         rel_path, pool),
+                                        pool);
+  return SVN_NO_ERROR;
+}
+
+
 
 /*** Fallback implementation of svn_ra_get_locations(). ***/
 
@@ -223,7 +250,7 @@ log_receiver(void *baton,
   const char *prev_path;
 
   /* No paths were changed in this revision.  Nothing to do. */
-  if (! log_entry->changed_paths)
+  if (! log_entry->changed_paths2)
     return SVN_NO_ERROR;
 
   /* If we've run off the end of the path's history, there's nothing
@@ -261,7 +288,7 @@ log_receiver(void *baton,
 
   /* Figure out at which repository path our object of interest lived
      in the previous revision. */
-  SVN_ERR(prev_log_path(&prev_path, NULL, NULL, log_entry->changed_paths,
+  SVN_ERR(prev_log_path(&prev_path, NULL, NULL, log_entry->changed_paths2,
                         current_path, lrb->kind, log_entry->revision, pool));
 
   /* Squirrel away our "next place to look" path (suffer the strcmp
@@ -280,7 +307,7 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
                            apr_hash_t **locations_p,
                            const char *path,
                            svn_revnum_t peg_revision,
-                           apr_array_header_t *location_revisions,
+                           const apr_array_header_t *location_revisions,
                            apr_pool_t *pool)
 {
   apr_hash_t *locations = apr_hash_make(pool);
@@ -288,20 +315,18 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
   apr_array_header_t *targets;
   svn_revnum_t youngest_requested, oldest_requested, youngest, oldest;
   svn_node_kind_t kind;
-  const char *root_url, *url, *rel_path;
+  const char *fs_path;
+  apr_array_header_t *sorted_location_revisions;
 
-  /* Fetch the repository root URL and relative path. */
-  SVN_ERR(svn_ra_get_repos_root2(session, &root_url, pool));
-  SVN_ERR(svn_ra_get_session_url(session, &url, pool));
-  url = svn_path_join(url, path, pool);
-  rel_path = svn_path_uri_decode(url + strlen(root_url), pool);
+  /* Fetch the absolute FS path associated with PATH. */
+  SVN_ERR(get_fs_path(&fs_path, session, path, pool));
 
   /* Sanity check: verify that the peg-object exists in repos. */
   SVN_ERR(svn_ra_check_path(session, path, peg_revision, &kind, pool));
   if (kind == svn_node_none)
     return svn_error_createf(SVN_ERR_FS_NOT_FOUND, NULL,
                              _("Path '%s' doesn't exist in revision %ld"),
-                             rel_path, peg_revision);
+                             fs_path, peg_revision);
 
   /* Easy out: no location revisions. */
   if (! location_revisions->nelts)
@@ -313,11 +338,11 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
   /* Figure out the youngest and oldest revs (amongst the set of
      requested revisions + the peg revision) so we can avoid
      unnecessary log parsing. */
-  qsort(location_revisions->elts, location_revisions->nelts,
-        location_revisions->elt_size, compare_revisions);
-  oldest_requested = APR_ARRAY_IDX(location_revisions, 0, svn_revnum_t);
-  youngest_requested = APR_ARRAY_IDX(location_revisions,
-                                     location_revisions->nelts - 1,
+  sorted_location_revisions = apr_array_copy(pool, location_revisions);
+  svn_sort__array(sorted_location_revisions, compare_revisions);
+  oldest_requested = APR_ARRAY_IDX(sorted_location_revisions, 0, svn_revnum_t);
+  youngest_requested = APR_ARRAY_IDX(sorted_location_revisions,
+                                     sorted_location_revisions->nelts - 1,
                                      svn_revnum_t);
   youngest = peg_revision;
   youngest = (oldest_requested > youngest) ? oldest_requested : youngest;
@@ -328,8 +353,8 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
 
   /* Populate most of our log receiver baton structure. */
   lrb.kind = kind;
-  lrb.last_path = rel_path;
-  lrb.location_revisions = apr_array_copy(pool, location_revisions);
+  lrb.last_path = fs_path;
+  lrb.location_revisions = apr_array_copy(pool, sorted_location_revisions);
   lrb.peg_revision = peg_revision;
   lrb.peg_path = NULL;
   lrb.locations = locations;
@@ -347,7 +372,7 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
 
   /* If the received log information did not cover any of the
      requested revisions, use the last known path.  (This normally
-     just means that ABS_PATH was not modified between the requested
+     just means that FS_PATH was not modified between the requested
      revision and OLDEST.  If the file was created at some point after
      OLDEST, then lrb.last_path should be NULL.) */
   if (! lrb.peg_path)
@@ -355,9 +380,9 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
   if (lrb.last_path)
     {
       int i;
-      for (i = 0; i < location_revisions->nelts; i++)
+      for (i = 0; i < sorted_location_revisions->nelts; i++)
         {
-          svn_revnum_t rev = APR_ARRAY_IDX(location_revisions, i,
+          svn_revnum_t rev = APR_ARRAY_IDX(sorted_location_revisions, i,
                                            svn_revnum_t);
           if (! apr_hash_get(locations, &rev, sizeof(rev)))
             apr_hash_set(locations, apr_pmemdup(pool, &rev, sizeof(rev)),
@@ -370,15 +395,15 @@ svn_ra__locations_from_log(svn_ra_session_t *session,
     return svn_error_createf
       (APR_EGENERAL, NULL,
        _("Unable to find repository location for '%s' in revision %ld"),
-       rel_path, peg_revision);
+       fs_path, peg_revision);
 
   /* Sanity check: make sure that our calculated peg path is the same
      as what we expected it to be. */
-  if (strcmp(rel_path, lrb.peg_path) != 0)
+  if (strcmp(fs_path, lrb.peg_path) != 0)
     return svn_error_createf
       (SVN_ERR_CLIENT_UNRELATED_RESOURCES, NULL,
        _("'%s' in revision %ld is an unrelated object"),
-       rel_path, youngest);
+       fs_path, youngest);
 
   *locations_p = locations;
   return SVN_NO_ERROR;
@@ -455,7 +480,7 @@ gls_log_receiver(void *baton,
      in the previous revision, and if its current location is the
      result of copy since then. */
   SVN_ERR(prev_log_path(&prev_path, NULL, &copyfrom_rev,
-                        log_entry->changed_paths, current_path,
+                        log_entry->changed_paths2, current_path,
                         lrb->kind, log_entry->revision, pool));
 
   /* If we've run off the end of the path's history, we need to report
@@ -512,13 +537,10 @@ svn_ra__location_segments_from_log(svn_ra_session_t *session,
   apr_array_header_t *targets;
   svn_node_kind_t kind;
   svn_revnum_t youngest_rev = SVN_INVALID_REVNUM;
-  const char *root_url, *url, *rel_path;
+  const char *fs_path;
 
-  /* Fetch the repository root URL and relative path. */
-  SVN_ERR(svn_ra_get_repos_root2(session, &root_url, pool));
-  SVN_ERR(svn_ra_get_session_url(session, &url, pool));
-  url = svn_path_join(url, path, pool);
-  rel_path = svn_path_uri_decode(url + strlen(root_url), pool);
+  /* Fetch the absolute FS path associated with PATH. */
+  SVN_ERR(get_fs_path(&fs_path, session, path, pool));
 
   /* If PEG_REVISION is invalid, it means HEAD.  If START_REV is
      invalid, it means HEAD.  If END_REV is SVN_INVALID_REVNUM, we'll
@@ -541,18 +563,18 @@ svn_ra__location_segments_from_log(svn_ra_session_t *session,
     }
 
   /* The API demands a certain ordering of our revision inputs. Enforce it. */
-  assert((peg_revision >= start_rev) && (start_rev >= end_rev));
+  SVN_ERR_ASSERT((peg_revision >= start_rev) && (start_rev >= end_rev));
 
   /* Sanity check: verify that the peg-object exists in repos. */
   SVN_ERR(svn_ra_check_path(session, path, peg_revision, &kind, pool));
   if (kind == svn_node_none)
     return svn_error_createf(SVN_ERR_FS_NOT_FOUND, NULL,
                              _("Path '%s' doesn't exist in revision %ld"),
-                             rel_path, start_rev);
+                             fs_path, start_rev);
 
   /* Populate most of our log receiver baton structure. */
   lrb.kind = kind;
-  lrb.last_path = rel_path;
+  lrb.last_path = fs_path;
   lrb.done = FALSE;
   lrb.start_rev = start_rev;
   lrb.range_end = start_rev;
@@ -611,7 +633,6 @@ fr_log_message_receiver(void *baton,
 {
   struct fr_log_message_baton *lmb = baton;
   struct rev *rev;
-  apr_hash_index_t *hi;
 
   rev = apr_palloc(lmb->pool, sizeof(*rev));
   rev->revision = log_entry->revision;
@@ -620,20 +641,10 @@ fr_log_message_receiver(void *baton,
   lmb->eldest = rev;
 
   /* Duplicate log_entry revprops into rev->props */
-  rev->props = apr_hash_make(lmb->pool);
-  for (hi = apr_hash_first(pool, log_entry->revprops); hi;
-       hi = apr_hash_next(hi))
-    {
-      void *val;
-      const void *key;
-
-      apr_hash_this(hi, &key, NULL, &val);
-      apr_hash_set(rev->props, apr_pstrdup(lmb->pool, key), APR_HASH_KEY_STRING,
-                   svn_string_dup(val, lmb->pool));
-    }
+  rev->props = svn_prop_hash_dup(log_entry->revprops, lmb->pool);
 
   return prev_log_path(&lmb->path, &lmb->action,
-                       &lmb->copyrev, log_entry->changed_paths,
+                       &lmb->copyrev, log_entry->changed_paths2,
                        lmb->path, svn_node_file, log_entry->revision,
                        lmb->pool);
 }
@@ -648,38 +659,27 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
                            apr_pool_t *pool)
 {
   svn_node_kind_t kind;
-  const char *repos_url;
-  const char *session_url;
-  const char *tmp;
-  char *repos_abs_path;
+  const char *repos_url, *session_url, *fs_path;
   apr_array_header_t *condensed_targets;
   struct fr_log_message_baton lmb;
   struct rev *rev;
   apr_hash_t *last_props;
-  const char *last_path;
   svn_stream_t *last_stream;
   apr_pool_t *currpool, *lastpool;
 
-  SVN_ERR(svn_ra_get_repos_root2(ra_session, &repos_url, pool));
-  SVN_ERR(svn_ra_get_session_url(ra_session, &session_url, pool));
-
-  /* Create the initial path, using the repos_url and session_url */
-  tmp = svn_path_is_child(repos_url, session_url, pool);
-  repos_abs_path = apr_palloc(pool, strlen(tmp) + 1);
-  repos_abs_path[0] = '/';
-  memcpy(repos_abs_path + 1, tmp, strlen(tmp));
+  /* Fetch the absolute FS path associated with PATH. */
+  SVN_ERR(get_fs_path(&fs_path, ra_session, path, pool));
 
   /* Check to make sure we're dealing with a file. */
-  SVN_ERR(svn_ra_check_path(ra_session, "", end, &kind, pool));
-
+  SVN_ERR(svn_ra_check_path(ra_session, path, end, &kind, pool));
   if (kind == svn_node_dir)
     return svn_error_createf(SVN_ERR_FS_NOT_FILE, NULL,
-                             _("'%s' is not a file"), repos_abs_path);
+                             _("'%s' is not a file"), fs_path);
 
   condensed_targets = apr_array_make(pool, 1, sizeof(const char *));
-  APR_ARRAY_PUSH(condensed_targets, const char *) = "";
+  APR_ARRAY_PUSH(condensed_targets, const char *) = path;
 
-  lmb.path = svn_path_uri_decode(repos_abs_path, pool);
+  lmb.path = fs_path;
   lmb.eldest = NULL;
   lmb.pool = pool;
 
@@ -694,6 +694,8 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
                           pool));
 
   /* Reparent the session while we go back through the history. */
+  SVN_ERR(svn_ra_get_session_url(ra_session, &session_url, pool));
+  SVN_ERR(svn_ra_get_repos_root2(ra_session, &repos_url, pool));
   SVN_ERR(svn_ra_reparent(ra_session, repos_url, pool));
 
   currpool = svn_pool_create(pool);
@@ -701,7 +703,6 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
 
   /* We want the first txdelta to be against the empty file. */
   last_props = apr_hash_make(lastpool);
-  last_path = NULL;
   last_stream = svn_stream_empty(lastpool);
 
   /* Walk the revision list in chronological order, downloading each fulltext,
@@ -712,7 +713,6 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
   for (rev = lmb.eldest; rev; rev = rev->next)
     {
       const char *temp_path;
-      const char *temp_dir;
       apr_pool_t *tmppool;
       apr_hash_t *props;
       apr_file_t *file;
@@ -726,16 +726,12 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
 
       /* Get the contents of the file from the repository, and put them in
          a temporary local file. */
-      SVN_ERR(svn_io_temp_dir(&temp_dir, currpool));
-      SVN_ERR(svn_io_open_unique_file2
-              (&file, &temp_path,
-               svn_path_join(temp_dir, "tmp", currpool), ".tmp",
-               svn_io_file_del_on_pool_cleanup, currpool));
-      stream = svn_stream_from_aprfile(file, currpool);
+      SVN_ERR(svn_stream_open_unique(&stream, &temp_path, NULL,
+                                     svn_io_file_del_on_pool_cleanup,
+                                     currpool, currpool));
       SVN_ERR(svn_ra_get_file(ra_session, rev->path + 1, rev->revision,
                               stream, NULL, &props, currpool));
       SVN_ERR(svn_stream_close(stream));
-      SVN_ERR(svn_io_file_close(file, currpool));
 
       /* Open up a stream to the local file. */
       SVN_ERR(svn_io_file_open(&file, temp_path, APR_READ, APR_OS_DEFAULT,
@@ -753,8 +749,9 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
       /* Compute and send delta if client asked for it. */
       if (delta_handler)
         {
-          /* Get the content delta. */
-          svn_txdelta(&delta_stream, last_stream, stream, lastpool);
+          /* Get the content delta. Don't calculate checksums as we don't
+           * use them. */
+          svn_txdelta2(&delta_stream, last_stream, stream, FALSE, lastpool);
 
           /* And send. */
           SVN_ERR(svn_txdelta_send_txstream(delta_stream, delta_handler,
@@ -766,15 +763,192 @@ svn_ra__file_revs_from_log(svn_ra_session_t *ra_session,
       currpool = lastpool;
       lastpool = tmppool;
 
-      svn_stream_close(last_stream);
+      SVN_ERR(svn_stream_close(last_stream));
       last_stream = stream;
       last_props = props;
     }
 
-  svn_stream_close(last_stream);
+  SVN_ERR(svn_stream_close(last_stream));
   svn_pool_destroy(currpool);
   svn_pool_destroy(lastpool);
 
   /* Reparent the session back to the original URL. */
   return svn_ra_reparent(ra_session, session_url, pool);
 }
+
+
+/*** Fallback implementation of svn_ra_get_deleted_rev(). ***/
+
+/* svn_ra_get_log2() receiver_baton for svn_ra__get_deleted_rev_from_log(). */
+typedef struct log_path_del_rev_t
+{
+  /* Absolute repository path. */
+  const char *path;
+
+  /* Revision PATH was first deleted or replaced. */
+  svn_revnum_t revision_deleted;
+} log_path_del_rev_t;
+
+/* A svn_log_entry_receiver_t callback for finding the revision
+   ((log_path_del_rev_t *)BATON)->PATH was first deleted or replaced.
+   Stores that revision in ((log_path_del_rev_t *)BATON)->REVISION_DELETED.
+ */
+static svn_error_t *
+log_path_del_receiver(void *baton,
+                      svn_log_entry_t *log_entry,
+                      apr_pool_t *pool)
+{
+  log_path_del_rev_t *b = baton;
+  apr_hash_index_t *hi;
+
+  /* No paths were changed in this revision.  Nothing to do. */
+  if (! log_entry->changed_paths2)
+    return SVN_NO_ERROR;
+
+  for (hi = apr_hash_first(pool, log_entry->changed_paths2);
+       hi != NULL;
+       hi = apr_hash_next(hi))
+    {
+      void *val;
+      char *path;
+      svn_log_changed_path_t *log_item;
+
+      apr_hash_this(hi, (void *) &path, NULL, &val);
+      log_item = val;
+      if (svn_path_compare_paths(b->path, path) == 0
+          && (log_item->action == 'D' || log_item->action == 'R'))
+        {
+          /* Found the first deletion or replacement, we are done. */
+          b->revision_deleted = log_entry->revision;
+          break;
+        }
+    }
+  return SVN_NO_ERROR;
+}
+
+svn_error_t *
+svn_ra__get_deleted_rev_from_log(svn_ra_session_t *session,
+                                 const char *rel_deleted_path,
+                                 svn_revnum_t peg_revision,
+                                 svn_revnum_t end_revision,
+                                 svn_revnum_t *revision_deleted,
+                                 apr_pool_t *pool)
+{
+  const char *fs_path;
+  log_path_del_rev_t log_path_deleted_baton;
+
+  /* Fetch the absolute FS path associated with PATH. */
+  SVN_ERR(get_fs_path(&fs_path, session, rel_deleted_path, pool));
+
+  if (!SVN_IS_VALID_REVNUM(peg_revision))
+    return svn_error_createf(SVN_ERR_CLIENT_BAD_REVISION, NULL,
+                             _("Invalid peg revision %ld"), peg_revision);
+  if (!SVN_IS_VALID_REVNUM(end_revision))
+    return svn_error_createf(SVN_ERR_CLIENT_BAD_REVISION, NULL,
+                             _("Invalid end revision %ld"), end_revision);
+  if (end_revision <= peg_revision)
+    return svn_error_create(SVN_ERR_CLIENT_BAD_REVISION, NULL,
+                            _("Peg revision must precede end revision"));
+
+  log_path_deleted_baton.path = fs_path;
+  log_path_deleted_baton.revision_deleted = SVN_INVALID_REVNUM;
+
+  /* Examine the logs of SESSION's URL to find when DELETED_PATH was first
+     deleted or replaced. */
+  SVN_ERR(svn_ra_get_log2(session, NULL, peg_revision, end_revision, 0,
+                          TRUE, TRUE, FALSE,
+                          apr_array_make(pool, 0, sizeof(char *)),
+                          log_path_del_receiver, &log_path_deleted_baton,
+                          pool));
+  *revision_deleted = log_path_deleted_baton.revision_deleted;
+  return SVN_NO_ERROR;
+}
+
+
+svn_error_t *
+svn_ra__get_inherited_props_walk(svn_ra_session_t *session,
+                                 const char *path,
+                                 svn_revnum_t revision,
+                                 apr_array_header_t **inherited_props,
+                                 apr_pool_t *result_pool,
+                                 apr_pool_t *scratch_pool)
+{
+  const char *repos_root_url;
+  const char *session_url;
+  const char *parent_url;
+  apr_pool_t *iterpool = svn_pool_create(scratch_pool);
+
+  *inherited_props =
+    apr_array_make(result_pool, 1, sizeof(svn_prop_inherited_item_t *));
+
+  /* Walk to the root of the repository getting inherited
+     props for PATH. */
+  SVN_ERR(svn_ra_get_repos_root2(session, &repos_root_url, scratch_pool));
+  SVN_ERR(svn_ra_get_session_url(session, &session_url, scratch_pool));
+  parent_url = session_url;
+
+  while (strcmp(repos_root_url, parent_url))
+    {
+      apr_hash_index_t *hi;
+      apr_hash_t *parent_props;
+      apr_hash_t *final_hash = apr_hash_make(result_pool);
+      svn_error_t *err;
+
+      svn_pool_clear(iterpool);
+      parent_url = svn_uri_dirname(parent_url, scratch_pool);
+      SVN_ERR(svn_ra_reparent(session, parent_url, iterpool));
+      err = session->vtable->get_dir(session, NULL, NULL,
+                                     &parent_props, "",
+                                     revision, SVN_DIRENT_ALL,
+                                     iterpool);
+
+      /* If the user doesn't have read access to a parent path then
+         skip, but allow them to inherit from further up. */
+      if (err)
+        {
+          if ((err->apr_err == SVN_ERR_RA_NOT_AUTHORIZED)
+              || (err->apr_err == SVN_ERR_RA_DAV_FORBIDDEN))
+            {
+              svn_error_clear(err);
+              continue;
+            }
+          else
+            {
+              return svn_error_trace(err);
+            }
+        }
+
+      for (hi = apr_hash_first(scratch_pool, parent_props);
+           hi;
+           hi = apr_hash_next(hi))
+        {
+          const char *name = apr_hash_this_key(hi);
+          apr_ssize_t klen = apr_hash_this_key_len(hi);
+          svn_string_t *value = apr_hash_this_val(hi);
+
+          if (svn_property_kind2(name) == svn_prop_regular_kind)
+            {
+              name = apr_pstrdup(result_pool, name);
+              value = svn_string_dup(value, result_pool);
+              apr_hash_set(final_hash, name, klen, value);
+            }
+        }
+
+      if (apr_hash_count(final_hash))
+        {
+          svn_prop_inherited_item_t *new_iprop =
+            apr_palloc(result_pool, sizeof(*new_iprop));
+          new_iprop->path_or_url = svn_uri_skip_ancestor(repos_root_url,
+                                                         parent_url,
+                                                         result_pool);
+          new_iprop->prop_hash = final_hash;
+          svn_sort__array_insert(*inherited_props, &new_iprop, 0);
+        }
+    }
+
+  /* Reparent session back to original URL. */
+  SVN_ERR(svn_ra_reparent(session, session_url, scratch_pool));
+
+  svn_pool_destroy(iterpool);
+  return SVN_NO_ERROR;
+}
