@@ -2,17 +2,22 @@
  * blame-cmd.c -- Display blame information
  *
  * ====================================================================
- * Copyright (c) 2000-2008 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -21,43 +26,49 @@
 
 #include "svn_client.h"
 #include "svn_error.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_pools.h"
+#include "svn_props.h"
 #include "svn_cmdline.h"
+#include "svn_sorts.h"
 #include "svn_xml.h"
 #include "svn_time.h"
 #include "cl.h"
 
 #include "svn_private_config.h"
 
-typedef struct
+typedef struct blame_baton_t
 {
   svn_cl__opt_state_t *opt_state;
   svn_stream_t *out;
   svn_stringbuf_t *sbuf;
+
+  int rev_maxlength;
 } blame_baton_t;
 
 
 /*** Code. ***/
 
-/* This implements the svn_client_blame_receiver2_t interface, printing
+/* This implements the svn_client_blame_receiver3_t interface, printing
    XML to stdout. */
 static svn_error_t *
 blame_receiver_xml(void *baton,
+                   svn_revnum_t start_revnum,
+                   svn_revnum_t end_revnum,
                    apr_int64_t line_no,
                    svn_revnum_t revision,
-                   const char *author,
-                   const char *date,
+                   apr_hash_t *rev_props,
                    svn_revnum_t merged_revision,
-                   const char *merged_author,
-                   const char *merged_date,
+                   apr_hash_t *merged_rev_props,
                    const char *merged_path,
                    const char *line,
+                   svn_boolean_t local_change,
                    apr_pool_t *pool)
 {
-  svn_cl__opt_state_t *opt_state =
-    ((blame_baton_t *) baton)->opt_state;
-  svn_stringbuf_t *sb = ((blame_baton_t *) baton)->sbuf;
+  blame_baton_t *bb = baton;
+  svn_cl__opt_state_t *opt_state = bb->opt_state;
+  svn_stringbuf_t *sb = bb->sbuf;
 
   /* "<entry ...>" */
   /* line_no is 0-based, but the rest of the world is probably Pascal
@@ -66,19 +77,28 @@ blame_receiver_xml(void *baton,
                         "line-number",
                         apr_psprintf(pool, "%" APR_INT64_T_FMT,
                                      line_no + 1),
-                        NULL);
+                        SVN_VA_NULL);
 
   if (SVN_IS_VALID_REVNUM(revision))
-    svn_cl__print_xml_commit(&sb, revision, author, date, pool);
+    svn_cl__print_xml_commit(&sb, revision,
+                             svn_prop_get_value(rev_props,
+                                                SVN_PROP_REVISION_AUTHOR),
+                             svn_prop_get_value(rev_props,
+                                                SVN_PROP_REVISION_DATE),
+                             pool);
 
   if (opt_state->use_merge_history && SVN_IS_VALID_REVNUM(merged_revision))
     {
       /* "<merged>" */
       svn_xml_make_open_tag(&sb, pool, svn_xml_normal, "merged",
-                            "path", merged_path, NULL);
+                            "path", merged_path, SVN_VA_NULL);
 
-      svn_cl__print_xml_commit(&sb, merged_revision, merged_author,
-                               merged_date, pool);
+      svn_cl__print_xml_commit(&sb, merged_revision,
+                             svn_prop_get_value(merged_rev_props,
+                                                SVN_PROP_REVISION_AUTHOR),
+                             svn_prop_get_value(merged_rev_props,
+                                                SVN_PROP_REVISION_DATE),
+                             pool);
 
       /* "</merged>" */
       svn_xml_make_close_tag(&sb, pool, "merged");
@@ -102,13 +122,16 @@ print_line_info(svn_stream_t *out,
                 const char *date,
                 const char *path,
                 svn_boolean_t verbose,
+                int rev_maxlength,
                 apr_pool_t *pool)
 {
   const char *time_utf8;
   const char *time_stdout;
-  const char *rev_str = SVN_IS_VALID_REVNUM(revision)
-    ? apr_psprintf(pool, "%6ld", revision)
-                        : "     -";
+  const char *rev_str;
+
+  rev_str = SVN_IS_VALID_REVNUM(revision)
+    ? apr_psprintf(pool, "%*ld", rev_maxlength, revision)
+    : apr_psprintf(pool, "%*s", rev_maxlength, "-");
 
   if (verbose)
     {
@@ -137,32 +160,48 @@ print_line_info(svn_stream_t *out,
     }
   else
     {
-      return svn_stream_printf(out, pool, "%s %10s ", rev_str,
+      return svn_stream_printf(out, pool, "%s %10.10s ", rev_str,
                                author ? author : "         -");
     }
 
   return SVN_NO_ERROR;
 }
 
-/* This implements the svn_client_blame_receiver2_t interface. */
+/* This implements the svn_client_blame_receiver3_t interface. */
 static svn_error_t *
 blame_receiver(void *baton,
+               svn_revnum_t start_revnum,
+               svn_revnum_t end_revnum,
                apr_int64_t line_no,
                svn_revnum_t revision,
-               const char *author,
-               const char *date,
+               apr_hash_t *rev_props,
                svn_revnum_t merged_revision,
-               const char *merged_author,
-               const char *merged_date,
+               apr_hash_t *merged_rev_props,
                const char *merged_path,
                const char *line,
+               svn_boolean_t local_change,
                apr_pool_t *pool)
 {
-  svn_cl__opt_state_t *opt_state =
-    ((blame_baton_t *) baton)->opt_state;
-  svn_stream_t *out = ((blame_baton_t *)baton)->out;
+  blame_baton_t *bb = baton;
+  svn_cl__opt_state_t *opt_state = bb->opt_state;
+  svn_stream_t *out = bb->out;
   svn_boolean_t use_merged = FALSE;
 
+  if (!bb->rev_maxlength)
+    {
+      svn_revnum_t max_revnum = MAX(start_revnum, end_revnum);
+      /* The standard column width for the revision number is 6 characters.
+         If the revision number can potentially be larger (i.e. if the end_revnum
+          is larger than 1000000), we increase the column width as needed. */
+
+      bb->rev_maxlength = 6;
+      while (max_revnum >= 1000000)
+        {
+          bb->rev_maxlength++;
+          max_revnum = max_revnum / 10;
+        }
+    }
+
   if (opt_state->use_merge_history)
     {
       /* Choose which revision to use.  If they aren't equal, prefer the
@@ -172,19 +211,31 @@ blame_receiver(void *baton,
          we may need to adjust this. */
       if (merged_revision < revision)
         {
-          svn_stream_printf(out, pool, "G ");
+          SVN_ERR(svn_stream_puts(out, "G "));
           use_merged = TRUE;
         }
       else
-        svn_stream_printf(out, pool, "  ");
+        SVN_ERR(svn_stream_puts(out, "  "));
     }
 
   if (use_merged)
-    SVN_ERR(print_line_info(out, merged_revision, merged_author, merged_date,
-                            merged_path, opt_state->verbose, pool));
+    SVN_ERR(print_line_info(out, merged_revision,
+                            svn_prop_get_value(merged_rev_props,
+                                               SVN_PROP_REVISION_AUTHOR),
+                            svn_prop_get_value(merged_rev_props,
+                                               SVN_PROP_REVISION_DATE),
+                            merged_path, opt_state->verbose,
+                            bb->rev_maxlength,
+                            pool));
   else
-    SVN_ERR(print_line_info(out, revision, author, date, NULL,
-                            opt_state->verbose, pool));
+    SVN_ERR(print_line_info(out, revision,
+                            svn_prop_get_value(rev_props,
+                                               SVN_PROP_REVISION_AUTHOR),
+                            svn_prop_get_value(rev_props,
+                                               SVN_PROP_REVISION_DATE),
+                            NULL, opt_state->verbose,
+                            bb->rev_maxlength,
+                            pool));
 
   return svn_stream_printf(out, pool, "%s%s", line, APR_EOL_STR);
 }
@@ -204,10 +255,11 @@ svn_cl__blame(apr_getopt_t *os,
   int i;
   svn_boolean_t end_revision_unspecified = FALSE;
   svn_diff_file_options_t *diff_options = svn_diff_file_options_create(pool);
+  svn_boolean_t seen_nonexistent_target = FALSE;
 
   SVN_ERR(svn_cl__args_to_target_array_print_reserved(&targets, os,
                                                       opt_state->targets,
-                                                      ctx, pool));
+                                                      ctx, FALSE, pool));
 
   /* Blame needs a file on which to operate. */
   if (! targets->nelts)
@@ -241,9 +293,10 @@ svn_cl__blame(apr_getopt_t *os,
   if (! opt_state->xml)
     SVN_ERR(svn_stream_for_stdout(&bl.out, pool));
   else
-    bl.sbuf = svn_stringbuf_create("", pool);
+    bl.sbuf = svn_stringbuf_create_empty(pool);
 
   bl.opt_state = opt_state;
+  bl.rev_maxlength = 0;
 
   subpool = svn_pool_create(pool);
 
@@ -280,7 +333,7 @@ svn_cl__blame(apr_getopt_t *os,
       const char *target = APR_ARRAY_IDX(targets, i, const char *);
       const char *truepath;
       svn_opt_revision_t peg_revision;
-      svn_client_blame_receiver2_t receiver;
+      svn_client_blame_receiver3_t receiver;
 
       svn_pool_clear(subpool);
       SVN_ERR(svn_cl__check_cancel(ctx->cancel_baton));
@@ -296,7 +349,7 @@ svn_cl__blame(apr_getopt_t *os,
           else if (svn_path_is_url(target))
             opt_state->end_revision.kind = svn_opt_revision_head;
           else
-            opt_state->end_revision.kind = svn_opt_revision_base;
+            opt_state->end_revision.kind = svn_opt_revision_working;
         }
 
       if (opt_state->xml)
@@ -306,16 +359,16 @@ svn_cl__blame(apr_getopt_t *os,
              a target element if this path is skipped. */
           const char *outpath = truepath;
           if (! svn_path_is_url(target))
-            outpath = svn_path_local_style(truepath, subpool);
+            outpath = svn_dirent_local_style(truepath, subpool);
           svn_xml_make_open_tag(&bl.sbuf, pool, svn_xml_normal, "target",
-                                "path", outpath, NULL);
+                                "path", outpath, SVN_VA_NULL);
 
           receiver = blame_receiver_xml;
         }
       else
         receiver = blame_receiver;
 
-      err = svn_client_blame4(truepath,
+      err = svn_client_blame5(truepath,
                               &peg_revision,
                               &opt_state->start_revision,
                               &opt_state->end_revision,
@@ -333,12 +386,24 @@ svn_cl__blame(apr_getopt_t *os,
             {
               svn_error_clear(err);
               SVN_ERR(svn_cmdline_fprintf(stderr, subpool,
-                                          _("Skipping binary file: '%s'\n"),
+                                          _("Skipping binary file "
+                                            "(use --force to treat as text): "
+                                            "'%s'\n"),
                                           target));
             }
+          else if (err->apr_err == SVN_ERR_WC_PATH_NOT_FOUND ||
+                   err->apr_err == SVN_ERR_ENTRY_NOT_FOUND ||
+                   err->apr_err == SVN_ERR_FS_NOT_FILE ||
+                   err->apr_err == SVN_ERR_FS_NOT_FOUND)
+            {
+              svn_handle_warning2(stderr, err, "svn: ");
+              svn_error_clear(err);
+              err = NULL;
+              seen_nonexistent_target = TRUE;
+            }
           else
             {
-              return err;
+              return svn_error_trace(err);
             }
         }
       else if (opt_state->xml)
@@ -355,5 +420,11 @@ svn_cl__blame(apr_getopt_t *os,
   if (opt_state->xml && ! opt_state->incremental)
     SVN_ERR(svn_cl__xml_print_footer("blame", pool));
 
-  return SVN_NO_ERROR;
+  if (seen_nonexistent_target)
+    return svn_error_create(
+      SVN_ERR_ILLEGAL_TARGET, NULL,
+      _("Could not perform blame on all targets because some "
+        "targets don't exist"));
+  else
+    return SVN_NO_ERROR;
 }
